@@ -1,12 +1,13 @@
 from __future__ import annotations
 import logging
 import math
+import time
 from datetime import datetime
 from enum import StrEnum, auto
-from typing import Iterable
+from typing import Iterable, Literal
 
 import openpectus.protocol.models as Mdl
-from pydantic import BaseModel
+from pydantic import BaseModel, ConfigDict, NonNegativeInt, Field
 
 logger = logging.getLogger(__name__)
 
@@ -126,6 +127,7 @@ class TagsInfo(BaseModel):
             current.value = tag_value.value
             current.value_formatted = tag_value.value_formatted
             current.tick_time = tag_value.tick_time
+            current.simulated = tag_value.simulated
             return False  # was updated
 
     def __str__(self) -> str:
@@ -165,9 +167,19 @@ class RunData(BaseModel):
 class ActiveUser(BaseModel):
     """ Represents a user looking at the frontend details page for a process unit  """
 
-    id: str  # sid from identity token, used to get profile photos from ms graph api
+    id: str  # oid from identity token or a made up id from frontend if Anon, used to get profile photos from ms graph api
     name: str  # Same value as emitted by openpectus.aggregator.auth.user_name
 
+class Contributor(BaseModel):
+    """ Represents a contributor to a process unit """
+
+    id: str | None  # oid from identity token, or None if Anon
+    name: str  # Same value as emitted by openpectus.aggregator.auth.user_name
+
+    # from https://github.com/pydantic/pydantic/issues/1303#issuecomment-599712964
+    def __hash__(self):
+        return hash((type(self),) + tuple(self.__dict__.values()))
+
 
 class EngineData:
     """ Data stored by aggregator for each connected engine. """
@@ -199,7 +211,7 @@ class EngineData:
         self.error_log: AggregatedErrorLog = AggregatedErrorLog.empty()
         self.method_state: MethodState = MethodState.empty()
         self.plot_configuration: PlotConfiguration = PlotConfiguration.empty()
-        self.contributors: set[str] = set()
+        self.contributors: set[Contributor] = set()
         self.active_users: dict[str, ActiveUser] = dict()
         self.required_roles: set[str] = set()
         self.hardware_str: str = hardware_str
@@ -238,3 +250,57 @@ class EngineData:
         self.error_log.clear()
         self.method_state = MethodState.empty()
         self.contributors = set()
+
+
+class NotificationScope(StrEnum):
+    PROCESS_UNITS_WITH_RUNS_IVE_CONTRIBUTED_TO = auto()
+    PROCESS_UNITS_I_HAVE_ACCESS_TO = auto()
+    SPECIFIC_PROCESS_UNITS = auto()
+
+class NotificationTopic(StrEnum):
+    RUN_START = auto()
+    RUN_STOP = auto()
+    RUN_PAUSE = auto()
+    BLOCK_START = auto()
+    NOTIFICATION_CMD = auto()
+    WATCH_TRIGGERED = auto()
+    NEW_CONTRIBUTOR = auto()
+    METHOD_ERROR = auto()
+    NETWORK_ERRORS = auto()
+
+class WebPushNotificationPreferences(BaseModel):
+    model_config = ConfigDict(from_attributes=True)
+    user_id: str
+    user_roles: set[str]
+    scope: NotificationScope
+    topics: set[NotificationTopic]
+    process_units: set[str]
+
+
+
+class WebPushAction(BaseModel):
+    action: Literal['navigate']
+    title: str
+    icon: str | None = None
+
+
+class WebPushData(BaseModel):
+    process_unit_id: str | None = None
+    contributor_id: str | None = None
+
+
+class WebPushNotification(BaseModel): # see https://developer.mozilla.org/en-US/docs/Web/API/Notification/Notification#parameters for more information
+    actions: list[WebPushAction] = Field(default_factory=list)  # buttons user can press
+    badge: str | None = None  # url for smaller image for e.g. the notifcation bar
+    body: str | None = None
+    data: WebPushData = Field(default_factory=WebPushData)  # arbitrary data the frontend can use for e.g. navigating when user clicks an action button
+    icon: str | None = "/assets/icons/icon-192x192.png"  # url
+    image: str | None = None  # url
+    renotify: bool | None = None  # if set to true, tag must also be set
+    requireInteraction: bool | None = None  # notification will automatically close after a time unless this is set to True
+    silent: bool | None = None  # if True, notification will not make sound or vibration
+    tag: str | None = None  # an id for the notification, used for renotify
+    timestamp: NonNegativeInt | None = Field(default_factory=lambda: int(time.time()*1000))  # unix timestamp in milliseconds
+    title: str
+    vibrate: list[NonNegativeInt] | None = None
+
