@@ -4,7 +4,7 @@ import logging
 
 from Levenshtein import ratio
 
-from openpectus.lang.exec.visitor import NodeVisitor
+from openpectus.lang.exec.visitor import NodeGenerator, NodeVisitor
 import openpectus.lang.model.ast as p
 from openpectus.lang.exec.tags import TagValueCollection
 from openpectus.lang.exec.units import are_comparable, get_compatible_unit_names
@@ -92,8 +92,12 @@ class AnalyzerVisitorBase(NodeVisitor):
         self.items.append(item)
 
     def analyze(self, n: p.ProgramNode):
-        """ Run analysis synchronously. """
-        for _ in self.visit(n):
+        """ Run analysis synchronously.
+
+        Note: This runs visit_ProgramNode(n) for ProgramNode, same way as interpreter. For interpretation,
+        this differs from visit(n) but for analysis it should not make a difference.
+        """
+        for _ in self.visit_ProgramNode(n):
             pass
 
 
@@ -136,7 +140,8 @@ class UnreachableCodeCheckAnalyzer(AnalyzerVisitorBase):
             self.method_end = node
         yield
 
-    def visit_Node(self, node):
+    def visit_Node(self, node) -> NodeGenerator:
+        yield from ()
         super().visit_Node(node)
         if self.method_end and not self.first_command_after_end:
             self.first_command_after_end = node
@@ -253,7 +258,8 @@ class IndentationCheckAnalyzer(AnalyzerVisitorBase):
             start=0,
         )
 
-    def visit_Node(self, node: p.Node):
+    def visit_Node(self, node: p.Node) -> NodeGenerator:
+        yield from ()
         if node.indent_error and not isinstance(node, p.WhitespaceNode):
             self.add_item(self.create_item(node))
 
@@ -276,7 +282,9 @@ class ThresholdCheckAnalyzer(AnalyzerVisitorBase):
         # avoid this issue.
         self.max_threshold_in_parent: dict[p.Node, p.Node] = dict()
 
-    def visit_Node(self, node: p.Node):
+    def visit_Node(self, node: p.Node) -> NodeGenerator:
+        yield from ()
+
         if not isinstance(node, p.ProgramNode) and node.parent is not None and node.threshold is not None:
             possibly_limiting_node = self.max_threshold_in_parent.get(node.parent, None)
             if possibly_limiting_node is None or (possibly_limiting_node.threshold and node.threshold > possibly_limiting_node.threshold):
@@ -353,7 +361,13 @@ class WhitespaceCheckAnalyzer(AnalyzerVisitorBase):
         super().analyze(n)
 
         for node in self.whitespace_nodes:
-            if node.parent is not None and node.position.line > node.parent._last_non_ws_line:
+            last_non_ws_line = -1
+            parent = node.parent
+            while parent is not None:
+                if parent._last_non_ws_line > last_non_ws_line:
+                    last_non_ws_line = parent._last_non_ws_line
+                parent = parent.parent
+            if node.position.line > last_non_ws_line:
                 node.has_only_trailing_whitespace = True
 
 
@@ -387,8 +401,8 @@ class ConditionCheckAnalyzer(AnalyzerVisitorBase):
         self.analyze_condition(node)
         return super().visit_AlarmNode(node)
 
-    def analyze_condition(self, node: p.WatchNode | p.AlarmNode):
-        if node.condition is None:
+    def analyze_condition(self, node: p.NodeWithCondition):
+        if node.tag_operator_value is None:
             self.add_item(AnalyzerItem(
                 "ConditionMissing",
                 "Condition missing",
@@ -399,8 +413,8 @@ class ConditionCheckAnalyzer(AnalyzerVisitorBase):
                 end=len(node.instruction_part) + 1000  # Valid way to express the whole line
             ))
             return
-        assert node.condition is not None
-        condition = node.condition
+        assert node.tag_operator_value is not None
+        condition = node.tag_operator_value
 
         tag_name = condition.tag_name
         if tag_name is None or tag_name.strip() == '':
@@ -424,8 +438,8 @@ class ConditionCheckAnalyzer(AnalyzerVisitorBase):
                         node,
                         AnalyzerItemType.ERROR,
                         f"Did you mean to type '{most_similar_tag}'?",
-                        start=node.condition.lhs_range.start.character,
-                        end=node.condition.lhs_range.end.character,
+                        start=node.tag_operator_value.stripped_lhs_range.start.character,
+                        end=node.tag_operator_value.stripped_lhs_range.end.character,
                         data=dict(type="fix-typo", fix=most_similar_tag)
                     ))
                     return
@@ -436,8 +450,8 @@ class ConditionCheckAnalyzer(AnalyzerVisitorBase):
                     node,
                     AnalyzerItemType.ERROR,
                     f"The tag name '{tag_name}' is not valid",
-                    start=node.condition.lhs_range.start.character,
-                    end=node.condition.lhs_range.end.character,
+                    start=node.tag_operator_value.stripped_lhs_range.start.character,
+                    end=node.tag_operator_value.stripped_lhs_range.end.character,
                 ))
                 return
 
@@ -470,8 +484,8 @@ class ConditionCheckAnalyzer(AnalyzerVisitorBase):
                 node,
                 AnalyzerItemType.ERROR,
                 f"Unit '{condition.tag_unit}' was specified for a tag with no unit",
-                start=node.condition.rhs_range.start.character,
-                end=node.condition.rhs_range.end.character
+                start=node.tag_operator_value.tag_unit_range.start.character,
+                end=node.tag_operator_value.tag_unit_range.end.character
             ))
             return
 
@@ -487,8 +501,8 @@ class ConditionCheckAnalyzer(AnalyzerVisitorBase):
                 node,
                 AnalyzerItemType.ERROR,
                 "The tag requires that a value is provided.",
-                start=node.condition.rhs_range.start.character,
-                end=node.condition.rhs_range.end.character
+                start=node.tag_operator_value.rhs_range.start.character,
+                end=node.tag_operator_value.rhs_range.end.character
             ))
             return
 
@@ -500,8 +514,8 @@ class ConditionCheckAnalyzer(AnalyzerVisitorBase):
                 node,
                 AnalyzerItemType.ERROR,
                 "The tag requires that a unit is provided." + valid_units_str,
-                start=node.condition.rhs_range.start.character,
-                end=node.condition.rhs_range.end.character
+                start=node.tag_operator_value.stripped_rhs_range.start.character,
+                end=node.tag_operator_value.stripped_rhs_range.end.character
             ))
             return
 
@@ -516,8 +530,8 @@ class ConditionCheckAnalyzer(AnalyzerVisitorBase):
                     node,
                     AnalyzerItemType.ERROR,
                     str(vex) + "." + valid_units_str,
-                    start=node.condition.rhs_range.start.character,
-                    end=node.condition.rhs_range.end.character
+                    start=node.tag_operator_value.tag_unit_range.start.character,
+                    end=node.tag_operator_value.tag_unit_range.end.character
                 ))
                 return
             if not comparable:
@@ -527,12 +541,244 @@ class ConditionCheckAnalyzer(AnalyzerVisitorBase):
                     node,
                     AnalyzerItemType.ERROR,
                     f"The tag unit '{tag.unit}' is not compatible with the provided unit '{condition.tag_unit}'." + valid_units_str,
-                    start=node.condition.range.start.character,
-                    end=node.condition.range.end.character
+                    start=node.tag_operator_value.stripped_range.start.character,
+                    end=node.tag_operator_value.stripped_range.end.character
                 ))
                 return
 
 
+class SimulateCheckAnalyzer(AnalyzerVisitorBase):
+    """
+    Checks same as ConditionCheckAnalyzer
+    """
+
+    def __init__(self, tags: TagValueCollection) -> None:
+        super().__init__()
+        self.tags = tags
+
+    def __str__(self) -> str:
+        items = [str(item) for item in self.items]
+        tags = [str(tag) for tag in self.tags]
+        return f'{self.__class__.__name__}(items={items}, tags={tags})'
+
+    def visit_SimulateOffNode(self, node: p.SimulateOffNode):
+        tag_name = node.arguments
+        if tag_name is None or tag_name == '':
+            self.add_item(AnalyzerItem(
+                "MissingTag",
+                "Missing tag",
+                node,
+                AnalyzerItemType.ERROR,
+                "Argument must start with a tag name"
+            ))
+        elif not self.tags.has(tag_name):
+            if len(tag_name) > 2 and self.tags.names:
+                similarity = {tag: ratio(tag_name, tag) for tag in self.tags.names}
+                most_similar_tag = max(similarity, key=similarity.get)  # type: ignore
+                if similarity[most_similar_tag] > 0.7:
+                    self.add_item(AnalyzerItem(
+                        "UndefinedTag",
+                        "Undefined tag",
+                        node,
+                        AnalyzerItemType.ERROR,
+                        f"Did you mean to type '{most_similar_tag}'?",
+                        start=node.stripped_arguments_range.start.character,
+                        end=node.stripped_arguments_range.end.character,
+                        data=dict(type="fix-typo", fix=most_similar_tag)
+                    ))
+            else:
+                self.add_item(AnalyzerItem(
+                    "UndefinedTag",
+                    "Undefined tag",
+                    node,
+                    AnalyzerItemType.ERROR,
+                    f"The tag name '{tag_name}' is not valid",
+                    start=node.stripped_arguments_range.start.character,
+                    end=node.stripped_arguments_range.end.character,
+                ))
+        return super().visit_SimulateOffNode(node)
+
+    def visit_SimulateNode(self, node: p.SimulateNode):
+        if node.tag_operator_value is None:
+            self.add_item(AnalyzerItem(
+                "AssignmentMissing",
+                "Assignment missing",
+                node,
+                AnalyzerItemType.ERROR,
+                "Assignment of a value to a tag is required",
+                start=len(node.instruction_part) + 1,
+                end=len(node.instruction_part) + 1000  # Valid way to express the whole line
+            ))
+            return super().visit_SimulateNode(node)
+        assert node.tag_operator_value is not None
+        condition = node.tag_operator_value
+
+        tag_name = condition.tag_name
+        if tag_name is None or tag_name.strip() == '':
+            self.add_item(AnalyzerItem(
+                "MissingTag",
+                "Missing tag",
+                node,
+                AnalyzerItemType.ERROR,
+                "Argument must start with a tag name"
+            ))
+            return super().visit_SimulateNode(node)
+
+        if not self.tags.has(tag_name):
+            if len(tag_name) > 2 and self.tags.names:
+                similarity = {tag: ratio(tag_name, tag) for tag in self.tags.names}
+                most_similar_tag = max(similarity, key=similarity.get)  # type: ignore
+                if similarity[most_similar_tag] > 0.7:
+                    self.add_item(AnalyzerItem(
+                        "UndefinedTag",
+                        "Undefined tag",
+                        node,
+                        AnalyzerItemType.ERROR,
+                        f"Did you mean to type '{most_similar_tag}'?",
+                        start=node.tag_operator_value.stripped_lhs_range.start.character,
+                        end=node.tag_operator_value.stripped_lhs_range.end.character,
+                        data=dict(type="fix-typo", fix=most_similar_tag)
+                    ))
+                    return super().visit_SimulateNode(node)
+            else:
+                self.add_item(AnalyzerItem(
+                    "UndefinedTag",
+                    "Undefined tag",
+                    node,
+                    AnalyzerItemType.ERROR,
+                    f"The tag name '{tag_name}' is not valid",
+                    start=node.tag_operator_value.stripped_lhs_range.start.character,
+                    end=node.tag_operator_value.stripped_lhs_range.end.character,
+                ))
+                return super().visit_SimulateNode(node)
+
+        if condition.op != "=":
+            self.add_item(AnalyzerItem(
+                "MissingOperator",
+                "Missing equals symbol",
+                node,
+                AnalyzerItemType.ERROR,
+                "The tag name must be followed by an equals symbol"
+            ))
+            return super().visit_SimulateNode(node)
+
+        if condition.rhs == "" or condition.tag_value == "":
+            self.add_item(AnalyzerItem(
+                "MissingValue",
+                "Missing value",
+                node,
+                AnalyzerItemType.ERROR,
+                "Equals symbol (=) must be followed by a value"
+            ))
+            return super().visit_SimulateNode(node)
+
+        tag = self.tags.get(tag_name)
+
+        if tag.unit is None and condition.tag_unit is not None:
+            self.add_item(AnalyzerItem(
+                "UnexpectedUnit",
+                "Unexpected tag unit",
+                node,
+                AnalyzerItemType.ERROR,
+                f"Unit '{condition.tag_unit}' was specified for a tag with no unit",
+                start=node.tag_operator_value.tag_unit_range.start.character,
+                end=node.tag_operator_value.tag_unit_range.end.character
+            ))
+            return super().visit_SimulateNode(node)
+
+        valid_units = get_compatible_unit_names(tag.unit)
+        valid_units_str = ""
+        if len(valid_units) > 0:
+            valid_units_str = " Suggested units: " + ", ".join(valid_units) + "."
+
+        if tag.unit is not None and condition.rhs and condition.rhs.strip() in valid_units and condition.tag_unit is None:
+            self.add_item(AnalyzerItem(
+                "MissingValue",
+                "Missing value",
+                node,
+                AnalyzerItemType.ERROR,
+                "The tag requires that a value is provided.",
+                start=node.tag_operator_value.stripped_rhs_range.start.character,
+                end=node.tag_operator_value.stripped_rhs_range.end.character
+            ))
+            return super().visit_SimulateNode(node)
+
+
+        if tag.unit is not None and condition.tag_unit is None:
+            self.add_item(AnalyzerItem(
+                "MissingUnit",
+                "Missing unit",
+                node,
+                AnalyzerItemType.ERROR,
+                "The tag requires that a unit is provided." + valid_units_str,
+                start=node.tag_operator_value.stripped_rhs_range.start.character,
+                end=node.tag_operator_value.stripped_rhs_range.end.character
+            ))
+            return super().visit_SimulateNode(node)
+
+        if tag.unit is not None:
+            assert condition.tag_unit is not None
+            try:
+                comparable = are_comparable(tag.unit, condition.tag_unit)
+            except ValueError as vex:
+                self.add_item(AnalyzerItem(
+                    "InvalidUnit",
+                    "Invalid unit",
+                    node,
+                    AnalyzerItemType.ERROR,
+                    str(vex) + "." + valid_units_str,
+                    start=node.tag_operator_value.tag_unit_range.start.character,
+                    end=node.tag_operator_value.tag_unit_range.end.character
+                ))
+                return super().visit_SimulateNode(node)
+            if not comparable:
+                self.add_item(AnalyzerItem(
+                    "IncompatibleUnits",
+                    "Incompatible units",
+                    node,
+                    AnalyzerItemType.ERROR,
+                    f"The tag unit '{tag.unit}' is not compatible with the provided unit '{condition.tag_unit}'." + valid_units_str,
+                    start=node.tag_operator_value.stripped_range.start.character,
+                    end=node.tag_operator_value.stripped_range.end.character
+                ))
+                return super().visit_SimulateNode(node)
+        return super().visit_SimulateNode(node)
+
+
+# Show blue squiggly lines notifying that condition is based on a simulated value.
+# Possibly uncomment this code at a later time. Right now it does not work
+# because UOD tags and their state are cached.
+"""
+    def visit_WatchNode(self, node: p.WatchNode):
+        if node.tag_operator_value and node.tag_operator_value.tag_name:
+            if self.tags.has(node.tag_operator_value.tag_name) and self.tags.get(node.tag_operator_value.tag_name).simulated:
+                self.add_item(AnalyzerItem(
+                    "SimulatedValue",
+                    "Tag has simulated value",
+                    node,
+                    AnalyzerItemType.INFO,
+                    f"The value of tag {node.tag_operator_value.tag_name} is simulated.",
+                    start=node.tag_operator_value.stripped_lhs_range.start.character,
+                    end=node.tag_operator_value.stripped_lhs_range.end.character
+                ))
+        return super().visit_WatchNode(node)
+
+    def visit_AlarmNode(self, node: p.AlarmNode):
+        if node.tag_operator_value and node.tag_operator_value.tag_name:
+            if self.tags.has(node.tag_operator_value.tag_name) and self.tags.get(node.tag_operator_value.tag_name).simulated:
+                self.add_item(AnalyzerItem(
+                    "SimulatedValue",
+                    "Tag has simulated value",
+                    node,
+                    AnalyzerItemType.INFO,
+                    f"The value of tag {node.tag_operator_value.tag_name} is simulated.",
+                    start=node.tag_operator_value.stripped_lhs_range.start.character,
+                    end=node.tag_operator_value.stripped_lhs_range.end.character
+                ))
+        return super().visit_AlarmNode(node)
+"""
+
+
 class CommandCheckAnalyzer(AnalyzerVisitorBase):
     """
     Checks:
@@ -639,7 +885,7 @@ class MacroCheckAnalyzer(AnalyzerVisitorBase):
         self.macros_called: list[p.MacroNode] = []
 
     def visit_CallMacroNode(self, node: p.CallMacroNode):
-        if node.name is None or node.name.strip() == "":
+        if node.name == "":
             self.add_item(AnalyzerItem(
                 "MacroCallNameInvalid",
                 "Invalid macro call",
@@ -647,8 +893,8 @@ class MacroCheckAnalyzer(AnalyzerVisitorBase):
                 AnalyzerItemType.ERROR,
                 "Call macro must refer to a Macro definition"
             ))
-        elif node.name and node.name.strip() not in self.macros:
-            similarity = {macro: ratio(node.name.strip(), macro) for macro in self.macros.keys()}
+        elif node.name not in self.macros:
+            similarity = {macro: ratio(node.name, macro) for macro in self.macros.keys()}
             most_similar_macro = max(similarity, key=similarity.get) if similarity else None  # type: ignore
             if most_similar_macro and similarity[most_similar_macro] > 0.7:
                 self.add_item(AnalyzerItem(
@@ -657,8 +903,8 @@ class MacroCheckAnalyzer(AnalyzerVisitorBase):
                     node,
                     AnalyzerItemType.ERROR,
                     f"Did you mean to type '{most_similar_macro}'?",
-                    start=node.arguments_range.start.character,
-                    end=node.arguments_range.end.character,
+                    start=node.stripped_arguments_range.start.character,
+                    end=node.stripped_arguments_range.end.character,
                     data=dict(type="fix-typo", fix=most_similar_macro)
                 ))
             else:
@@ -667,27 +913,18 @@ class MacroCheckAnalyzer(AnalyzerVisitorBase):
                     "Referenced macro is not defined",
                     node,
                     AnalyzerItemType.ERROR,
-                    f"Cannot call macro '{node.name.strip()}' because it is not defined",
-                    start=node.arguments_range.start.character,
-                    end=node.arguments_range.end.character,
+                    f"Cannot call macro '{node.name}' because it is not defined",
+                    start=node.stripped_arguments_range.start.character,
+                    end=node.stripped_arguments_range.end.character,
                 ))
-        elif node.name and node.name.strip() in self.macros:
+        elif node.name in self.macros:
             macro_node = self.macros[node.name]
             self.macro_calls.append(node)
-            if macro_node not in node.parents:
-                self.macros_called.append(macro_node)
-            else:
-                self.add_item(AnalyzerItem(
-                    "MacroRecursive",
-                    "Macro calls itself",
-                    macro_node,
-                    AnalyzerItemType.ERROR,
-                    f"Macro are not allowed to contain calls to themselves (line {node.position.line+1})",
-                ))
+            self.macros_called.append(macro_node)
         return super().visit_CallMacroNode(node)
 
     def visit_MacroNode(self, node: p.MacroNode):
-        if node.name is None or node.name.strip() == "":
+        if node.name == "":
             self.add_item(AnalyzerItem(
                 "MacroNameInvalid",
                 "Invalid macro definition",
@@ -695,9 +932,9 @@ class MacroCheckAnalyzer(AnalyzerVisitorBase):
                 AnalyzerItemType.ERROR,
                 "A Macro definition must include a name"
             ))
-        elif node.name and node.name.strip():
-            if node.name.strip() in self.macros:
-                previous_macro_with_same_name = self.macros[node.name.strip()]
+        else:
+            if node.name in self.macros:
+                previous_macro_with_same_name = self.macros[node.name]
                 self.add_item(AnalyzerItem(
                     "MacroRedefined",
                     "Macro redefined",
@@ -705,8 +942,29 @@ class MacroCheckAnalyzer(AnalyzerVisitorBase):
                     AnalyzerItemType.INFO,
                     f"This macro definition overwrites the previous definition at line {previous_macro_with_same_name.position.line+1}."
                 ))
-            self.macros[node.name.strip()] = node
+            self.macros[node.name] = node
             self.macros_registered.append(node)
+
+            cascade = node.macro_calling_macro(self.macros)
+            if cascade and node.name in cascade:
+                if len(cascade) == 1:
+                    self.add_item(AnalyzerItem(
+                        "MacroRecursive",
+                        "Macro calls itself",
+                        node,
+                        AnalyzerItemType.ERROR,
+                        f'Macro "{node.name}" calls itself. Unfortunately, this is not allowed.'
+                    ))
+                elif len(cascade) > 1:
+                    path = " which calls ".join(f'macro "{link}"' for link in cascade)
+                    self.add_item(AnalyzerItem(
+                        "MacroRecursive",
+                        "Macro calls itself indirectly",
+                        node,
+                        AnalyzerItemType.ERROR,
+                        f'Macro "{node.name}" calls itself by calling {path}. Unfortunately, this is not allowed.'
+                    ))
+
         return super().visit_MacroNode(node)
 
     def analyze(self, n):
@@ -737,6 +995,7 @@ class SemanticCheckAnalyzer:
             ThresholdCheckAnalyzer(),
             WhitespaceCheckAnalyzer(),
             ConditionCheckAnalyzer(tags),
+            SimulateCheckAnalyzer(tags),
             CommandCheckAnalyzer(commands),
             MacroCheckAnalyzer(),
         ]
