@@ -1,6 +1,6 @@
 from __future__ import annotations
 from logging import Logger
-from typing import Self, Type, TypeVar, TypedDict
+from typing import Callable, Self, Type, TypeVar, TypedDict
 
 
 class Position:
@@ -10,9 +10,6 @@ class Position:
         self.character: int = character
         """ Character number on line, zero-based. Corresponds to the indentation of the line """
 
-    def is_empty(self) -> bool:
-        return self == Position.empty
-
     @staticmethod
     def empty() -> Position:
         return Position(line=-1, character=-1)
@@ -105,6 +102,7 @@ class NodeState(TypedDict):
     name: str
     started: bool
     completed: bool
+    failed: bool
     cancelled: bool
     forced: bool
     action_history: list[str]
@@ -131,9 +129,12 @@ class SupportCancelForce:
         """ Whether the node has been cancelled. Virtual property. """
         return self._cancelled
 
-    def cancel(self):
+    def cancel(self) -> bool:
+        """ Cancel the instruction. Returns a bool indication whether Cancel was valid at the time """
         if self.cancellable:
             self._cancelled = True
+            return True
+        return False
 
     @property
     def forcible(self) -> bool:
@@ -146,8 +147,11 @@ class SupportCancelForce:
         return self._forced
 
     def force(self):
+        """ Force the instruction. Returns a bool indication whether Force was valid at the time """
         if self.forcible:
             self._forced = True
+            return True
+        return False
 
 
 class SupportsInterrupt():
@@ -164,7 +168,7 @@ class Node(SupportCancelForce):
     instruction_names: list[str] = []
     """ Specifies which node the parser should instantiate for a given instruction name(s) """
 
-    def __init__(self, position=Position.empty, id=""):
+    def __init__(self, position: Position = Position.empty(), id=""):
         super().__init__()
 
         self.parent: NodeWithChildren | None = None
@@ -190,6 +194,7 @@ class Node(SupportCancelForce):
         # interpretation state
         self.started: bool = False
         self.completed: bool = False
+        self.failed: bool = False
         self.action_history: list[str] = []
         self._empty_node_class_names = [BlankNode.__name__, CommentNode.__name__]
 
@@ -206,10 +211,7 @@ class Node(SupportCancelForce):
 
     @property
     def runlog_name(self) -> str | None:
-        return self.display_name
-
-    @property
-    def display_name(self) -> str:
+        """ The name to display in the runlog. Includes argument, if any, but no threshold or comment"""
         args = ": " + self.arguments if len(self.arguments) > 0 else ""
         return self.instruction_name + args
 
@@ -239,6 +241,7 @@ class Node(SupportCancelForce):
             raise ValueError(f"Cannot load state from {state['class_name']} into {self.__class__}")
         self.started = state["started"]
         self.completed = state["completed"]
+        self.failed = state["failed"]
         self._cancelled = state["cancelled"]
         self._forced = state["forced"]
         self.action_history = state["action_history"]
@@ -250,6 +253,7 @@ class Node(SupportCancelForce):
             name=self.name,
             started=self.started,
             completed=self.completed,
+            failed=self.failed,
             cancelled=self.cancelled,
             forced=self.forced,
             action_history=self.action_history
@@ -270,6 +274,18 @@ class Node(SupportCancelForce):
 
         return get_by_instruction(self, instruction_name, arguments)
 
+    def get_child_by_condition(self, predicate: Callable[[Node], bool], include_self=True) -> Node | None:
+        """ Return the first child node that matches the condition """
+        if include_self and predicate(self):
+            return self
+        if isinstance(self, NodeWithChildren):
+            for child in self.children:
+                # Note: we must use include_self=True here regardless of its value above
+                result = child.get_child_by_condition(predicate, include_self=True)
+                if result is not None:
+                    return result
+        return None
+
     def with_id(self, gen: NodeIdGenerator) -> Self:
         self.id = gen.create_id(self)
         return self
@@ -292,6 +308,18 @@ class Node(SupportCancelForce):
         args = "" if self.arguments_part == "" else ": " + self.arguments_part
         return f"{indent_spaces}{self.instruction_part}{args} | id={self.id}"
 
+    @property
+    def class_name(self) -> str:
+        return self.__class__.__name__
+
+    @classmethod
+    def is_class_of(cls, node: Node | None = None) -> bool:
+        return node.__class__ == cls
+
+    @classmethod
+    def is_class_of_name(cls, node_class_name: str) -> bool:
+        return node_class_name == cls.__name__
+
     @property
     def parents(self) -> list[NodeWithChildren]:
         node = self
@@ -379,7 +407,7 @@ class Node(SupportCancelForce):
         return matches(self, other, logger)
 
 class NodeWithChildren(Node):
-    def __init__(self, position=Position.empty, id=""):
+    def __init__(self, position=Position.empty(), id=""):
         super().__init__(position, id)
         self._children: list[Node] = []
         self.interrupt_registered: bool = False
@@ -451,7 +479,7 @@ class NodeWithChildren(Node):
         super().reset_runtime_state(recursive)
 
 class ProgramNode(NodeWithChildren):
-    def __init__(self, position=Position.empty, id=""):
+    def __init__(self, position=Position.empty(), id=""):
         super().__init__(position, id)
         self.active_node: Node | None = None
         """ The node currently executing. Is never ProgramNode. Is None untli first instruction is
@@ -544,7 +572,7 @@ class ProgramNode(NodeWithChildren):
 class MarkNode(Node):
     instruction_names = ["Mark"]
 
-    def __init__(self, position=Position.empty, id=""):
+    def __init__(self, position=Position.empty(), id=""):
         super().__init__(position, id)
         self._forcible = False
 
@@ -552,7 +580,7 @@ class MarkNode(Node):
 class BlockNode(NodeWithChildren):
     instruction_names = ["Block"]
 
-    def __init__(self, position=Position.empty, id=""):
+    def __init__(self, position=Position.empty(), id=""):
         super().__init__(position, id)
         self.lock_aquired = False
 
@@ -585,7 +613,7 @@ class BatchNode(Node):
 class NodeWithTagOperatorValue(Node):
     operators: list[str]
 
-    def __init__(self, position=Position.empty, id=""):
+    def __init__(self, position=Position.empty(), id=""):
         super().__init__(position, id)
         self.tag_operator_value_part: str
         self.tag_operator_value: TagOperatorValue | None
@@ -594,7 +622,7 @@ class NodeWithTagOperatorValue(Node):
 class NodeWithCondition(NodeWithTagOperatorValue):
     operators = ["<=", ">=", "==", "!=", "<", ">", "="]
 
-    def __init__(self, position=Position.empty, id=""):
+    def __init__(self, position=Position.empty(), id=""):
         super().__init__(position, id)
         self.interrupt_registered: bool = False
         self.activated: bool = False
@@ -636,14 +664,32 @@ class NodeWithAssignment(NodeWithTagOperatorValue):
 class WatchNode(NodeWithChildren, NodeWithCondition, SupportsInterrupt):
     instruction_names = ["Watch"]
 
+    def __init__(self, position=Position.empty(), id=""):
+        super().__init__(position, id)
+        self.awaiting_condition = False
+
+    def extract_state(self):
+        state = super().extract_state()
+        state["awaiting_condition"] = self.awaiting_condition  # type: ignore
+        return state
+
+    def apply_state(self, state):
+        self.awaiting_condition = bool(state["awaiting_condition"])
+        super().apply_state(state)
+
+    def reset_runtime_state(self, recursive):
+        self.awaiting_condition = False
+        super().reset_runtime_state(recursive)
+
 
 class AlarmNode(NodeWithChildren, NodeWithCondition, SupportsInterrupt):
     instruction_names = ["Alarm"]
 
-    def __init__(self, position=Position.empty, id=""):
+    def __init__(self, position=Position.empty(), id=""):
         super().__init__(position, id)
         self.run_count: int = 0
         """ The number of times the alarm has completed """
+        self.awaiting_condition = False
 
     def extract_state(self):
         state = super().extract_state()
@@ -656,6 +702,7 @@ class AlarmNode(NodeWithChildren, NodeWithCondition, SupportsInterrupt):
 
     def reset_runtime_state(self, recursive):
         # Note: run_count is not reset because it counts alarm invocations
+        self.awaiting_condition = False
         super().reset_runtime_state(recursive)
 
 
@@ -691,7 +738,7 @@ class WhitespaceNode(Node):
 
     Populated by WhitespaceAnalyzer
     """
-    def __init__(self, position=Position.empty, id=""):
+    def __init__(self, position=Position.empty(), id=""):
         super().__init__(position, id)
         self.has_only_trailing_whitespace: bool = False
         """ Specifies that only whitespace instructions follow this whitespace instruction in
@@ -700,7 +747,7 @@ class WhitespaceNode(Node):
 
 class CommentNode(WhitespaceNode):
     """ Represents a line with only a comment. """
-    def __init__(self, position=Position.empty, id=""):
+    def __init__(self, position=Position.empty(), id=""):
         super().__init__(position, id)
         self.has_comment = True
         self.line: str = ""
@@ -717,7 +764,7 @@ class InjectedNode(NodeWithChildren, SupportsInterrupt):
 class MacroNode(NodeWithChildren):
     instruction_names = ["Macro"]
 
-    def __init__(self, position=Position.empty, id=""):
+    def __init__(self, position=Position.empty(), id=""):
         super().__init__(position, id)
         self.activated: bool = False
         self._cancellable = False
@@ -774,7 +821,7 @@ class MacroNode(NodeWithChildren):
 class CallMacroNode(Node):
     instruction_names = ["Call macro"]
 
-    def __init__(self, position=Position.empty, id=""):
+    def __init__(self, position=Position.empty(), id=""):
         super().__init__(position, id)
         self._cancellable = False
         self._forcible = False
@@ -797,7 +844,7 @@ class CallMacroNode(Node):
 class NotifyNode(Node):
     instruction_names = ["Notify"]
 
-    def __init__(self, position=Position.empty, id=""):
+    def __init__(self, position=Position.empty(), id=""):
         super().__init__(position, id)
         self._cancellable = False
         self._forcible = False
@@ -811,7 +858,7 @@ class InterpreterCommandNode(CommandBaseNode):
     """ Represents commands that are directly executable by the interpreter. """
     instruction_names = ["Base", "Increment run counter", "Run counter", "Wait"]
 
-    def __init__(self, position=Position.empty, id=""):
+    def __init__(self, position=Position.empty(), id=""):
         super().__init__(position, id)
         self.wait_start_time: float | None = None
 
@@ -845,6 +892,9 @@ class SimulateOffNode(Node):
 
 class UodCommandNode(CommandBaseNode):
     """ Represents a uod command, subclassing UodCommand. """
+    def __init__(self, position=Position.empty(), id=""):
+        super().__init__(position, id)
+        self._cancellable = True
 
 
 class Comment:
@@ -860,17 +910,38 @@ class Error:
 
 class BlankNode(WhitespaceNode):
     """ Represents a line that contains only whitespace. """
-    def __init__(self, position=Position.empty, id=""):
+    def __init__(self, position=Position.empty(), id=""):
         super().__init__(position, id)
 
 
 class ErrorInstructionNode(Node):
     """ Represents non-parsable instruction line. """
 
-    def __init__(self, position=Position.empty, id=""):
+    def __init__(self, position=Position.empty(), id=""):
         super().__init__(position, id)
         self.line: str = ""
 
     def with_line(self, line: str):
         self.line = line
         return self
+
+
+class NullNode(Node):
+    """ Marker type for temporary nodes, notably for Start which is executed before tracking is initialized. """
+    def __init__(self, id: str, command_name: str):
+        self._command_name = command_name
+        self.arguments = command_name  # make runlog aware of name
+        super().__init__(Position.empty(), id)
+
+    @property
+    def command_name(self) -> str:
+        """ Name of the original command """
+        return self._command_name
+
+    @property
+    def runlog_name(self) -> str | None:
+        return "NullNode"
+
+    def __str__(self):
+        return f"{self.__class__.__name__}(command_name='{self.command_name}', " + \
+            f"arguments={self.arguments}, id='{self.id}')"
