@@ -3,8 +3,10 @@ import time
 import unittest
 from typing import Any
 
-from openpectus.lang.exec.errors import EngineError, MethodEditError
+from openpectus.engine.models import EngineCommandEnum
+from openpectus.lang.exec.errors import EngineError, InterpretationError, MethodEditError
 from openpectus.lang.exec.regex import RegexNumber
+from openpectus.lang.exec.runlog import RunLogItemState
 from openpectus.lang.exec.tags_impl import ReadingTag, SelectTag
 from openpectus.engine.hardware import RegisterDirection
 
@@ -19,7 +21,9 @@ from openpectus.lang.model.parser import ParserMethod, create_method_parser
 from openpectus.protocol.models import Method
 from openpectus.test.engine.utility_methods import (
     EngineTestRunner,
-    configure_test_logger, set_engine_debug_logging, set_interpreter_debug_logging
+    configure_test_logger,
+    print_runlog,
+    print_runtime_records, set_engine_debug_logging, set_interpreter_debug_logging
 )
 import openpectus.lang.model.ast as p
 
@@ -144,7 +148,6 @@ class TestMethodManager(unittest.TestCase):
             with self.assertRaises(MethodEditError):
                 instance.engine.set_method(method2)
 
-
     def test_may_edit_line_awaiting_threshold(self):
 
         method1 = Method.from_numbered_pcode("""\
@@ -908,11 +911,34 @@ Watch: Run counter > 0
 
 # End Generational Edits
 
-    @unittest.skip("TODO")
     def test_edit_injected(self):
-        # hmm this is weird. a method is running and user injects code - that's not a method edit, just an injection
-        # discussion in issue #829
-        raise NotImplementedError()
+        method1 = Method.from_numbered_pcode("""\
+01 Mark: A
+02 Mark: B
+03 Mark: C
+04 
+""")
+        method2 = Method.from_numbered_pcode("""\
+01 Mark: A
+02 Mark: B
+03 Mark: C
+04 Mark: D
+""")
+        runner = EngineTestRunner(create_test_uod, method1)
+        with runner.run() as instance:
+            instance.start()
+
+            # method1 runs - macro registered
+            instance.run_until_instruction("Mark", state="completed", arguments="A")
+            instance.engine.inject_code("Mark: A2")
+            instance.run_until_instruction("Mark", state="completed", arguments="C")
+
+            self.assertEqual(['A', 'B', 'A2', 'C'], instance.marks)
+
+            instance.engine.set_method(method2)
+            instance.run_until_instruction("Mark", state="completed", arguments="D")
+
+            self.assertEqual(['A', 'B', 'A2', 'C', 'D'], instance.marks)
 
     def test_active_node(self):
         # program.active_node is never ProgramNode and is only None right at the beginning until the first program child
@@ -1002,7 +1028,7 @@ Watch: Run counter > 0
         edit_at(7)
 
 
-    def test_command_exec_id(self):
+    def test_command_instance_id(self):
         # Check how it works if a program containing commands is edited.
         # Specifically, RuntimeInfo.with_edited_program() use RuntimeRecordState.clone() which reuses the
         # command instance and command_exec_id which may be a problem.
@@ -1029,12 +1055,19 @@ Watch: Run counter > 0
             # Wait until mid-execution of Reset's 5 ticks, note that this is right at the end of the v0 method
             instance.run_until_instruction("Wait", state="completed", arguments="0.2s")
 
+        #     print_runtime_records(instance.engine, "pre-edit")
+        #     print_runlog(instance.engine, "pre-edit")
+
             instance.engine.set_method(method2)
 
+        #     instance.run_ticks(1)
+        #     print_runtime_records(instance.engine, "post-edit")
+        #     print_runlog(instance.engine, "post-edit")
+
             # note how Reset still gets ticked by engine, even though it's water under the bridge for interpreter
             instance.run_until_instruction("Wait", state="completed", arguments="0.6s")
 
-    def test_command_exec_id_2(self):
+    def test_command_instance_id_2(self):
         # Variation of the above that performs the edit earlier than end-of-method
 
         method1 = Method.from_numbered_pcode("""\
@@ -1090,3 +1123,53 @@ Watch: Run counter > 0
             self.assertEqual(0, instance.method_manager.program.revision)
             instance.run_until_instruction("Mark", arguments="C")
             self.assertEqual(['A', 'B', 'C'], instance.marks)
+
+# Resume after error
+
+    def test_resume_after_error_invalid_unit(self):
+
+        method_w_error = Method.from_numbered_pcode("""\
+00 Info: Start
+01 Watch: Run Time > 1x
+02     Mark: A
+""")
+        runner = EngineTestRunner(create_test_uod, method_w_error)
+        with runner.run() as instance:
+            instance.start()
+
+            with self.assertRaises(EngineError) as ctx:
+                instance.run_until_instruction("Watch", "completed")
+
+            ex = ctx.exception.__cause__
+            assert isinstance(ex, InterpretationError)
+            assert "Invalid unit: 'x'" in ex.message
+            watch_node = instance.method_manager.program.get_child_by_id("01")
+            assert watch_node is not None
+            assert watch_node.failed
+            assert instance.engine.has_error_state()
+
+            print_runtime_records(instance.engine, "Error has just occurred")
+
+            method_corrected = Method.from_numbered_pcode("""\
+00 Info: Start
+01 Watch: Run Time > 1s
+02     Mark: A
+""")
+            # User edits and clicks Save
+            instance.engine.set_method(method_corrected)
+
+            print_runtime_records(instance.engine, "Method corrected")
+
+            assert not instance.engine.has_error_state()
+            watch_node = instance.method_manager.program.get_child_by_id("01")
+            assert watch_node is not None
+            assert watch_node.arguments == "Run Time > 1s"
+            assert not watch_node.failed
+
+            # user clicks Pause to resume (Unpause) from the error instruction
+            instance.engine.schedule_execution(EngineCommandEnum.UNPAUSE)
+
+            instance.run_until_instruction("Watch", "completed")
+            self.assertEqual(['A'], instance.marks)
+
+# End Resume after error
